from autoprotocol import UserError
from collections import namedtuple
from autoprotocol.container import Well, WellGroup
import datetime
import sys

if sys.version_info[0] >= 3:
    string_type = str
else:
    string_type = basestring


def user_errors_group(error_msgs, info=None):
    """Takes a list error messages and neatly displays as a single UserError

    Will automatically remove instances of None and only report errors if list
    is not empty.

    Parameters
    ----------
    error_msgs : list
        List of strings that are the error messages

    Raises
    ------
    ValueError
        If error_msgs is not of type list

    """
    assert isinstance(error_msgs, list), ("Error messages must be in the form"
                                          " of a list to properly format the "
                                          "grouped message.")

    error_msgs = [_f for _f in error_msgs if _f]
    if len(error_msgs) != 0:
        raise UserError(
            "%s error(s) found in this protocol: " % len(error_msgs) +
            " ".join(["<Error " +
                      str(i + 1) + "> " +
                      str(m) for i, m in enumerate(error_msgs)]), info=info)


def printdatetime():
    """
    Generate a datetime string.

    Returns
    -------
    str
        The current date and time formatted as: YYYY-MM-DD_HH-MM-SS

    """
    printdate = datetime.datetime.now().strftime('%Y-%m-%d_%H:%M:%S')
    return printdate


def printdate():
    """
    Generate a date string.

    Returns
    -------
    str
        The current date formatted as: YYYY-MM-DD

    """
    printdate = datetime.datetime.now().strftime('%Y-%m-%d')
    return printdate


def make_list(my_str, integer=False):
    """
    Return a list of strings (or ints) from a string containing
    comma separated elements.

    Parameters
    ----------
    my_str : str
        String with individual elements separated by comma.
    interger : bool
        If true list of integers instead of list of strings
        is returned. Default: False.

    Returns
    -------
    list
        List of strings or integers

    Raises
    ------
    ValueError
        If my_str is not of type string

    """
    assert isinstance(my_str, string_type), (
        "Input needs to be of type string")
    if integer:
        my_str = [int(x.strip()) for x in my_str.split(",")]
    else:
        my_str = [x.strip() for x in my_str.split(",")]
    return my_str


def flatten_list(l):
    """
    Flatten a list recursively without for loops or additional modules

    Example Usage:

    .. code-block:: python

            from autoprotocol_utilities.misc_helpers import flatten_list

            l = [-1, 0, [1,2], "string", [3, [4, 5]]]
            flatten_list(l)

    Returns:

    .. code-block:: python

            [-1, 0, 1, 2, 'string', 3, 4, 5]

    Parameters
    ---------
    l : list, list of WellGroup
        List or list of WellGroups to flatten


    Returns
    -------
    list
        Flat list

    Raises
    ------
    ValueError
        If l is not of type list

    """
    if isinstance(l, (list, WellGroup)):
        return [x for sublist in l for x in flatten_list(sublist)]
    else:
        return [l]


def det_new_group(i, base=0):
    """Determine if new_group should be added to pipetting operation.

    Helper to determine if new_group should be added. Returns true when
    'i' matches the base, which defaults to 0.

    Parameters
    ----------
    i : int
        The iteration you are on.
    base : int, optional
        The value at which you want to trigger a new group. Default: 0.

    Returns
    -------
    bool
        True if iteration equals base case.

    Raises
    ------
    ValueError
        If i is not of type integer
    ValueError
        If base is not of type integer

    """
    assert isinstance(i, int), "Needs an integer."
    assert isinstance(base, int), "Base has to be an integer"
    if i == base:
        new_group = True
    else:
        new_group = False
    return new_group


def char_limit(label, length=22, trunc=False, clip=False):
    """Enforces a string limit on the label provided

    Can either throw an error message if `length` is exceeded or correct the
    string and return the corrected string.

    Parameters
    ----------
    label : str
        String to test.
    length : int, optional
        Maximum label length for this string. Default: 22.
    trunc : bool, optional
        Truncate the label if it is too long. Default: False.
    clip : bool, optional
        Clip the label (remove from beginning of the string) if it is too long
        Default: False. If both trunc and clip are True, trunc will take effect
        and not clip.

    Returns
    -------
    namedtuple
        `label` (str) and `error_message` (string) that is empty on success
        `label` is the unmodified, truncated to clipped label as indicated

    Raises
    ------
    ValueError
        If label is not of type string

    """
    assert isinstance(label, string_type), "Label has to be of type string"

    r = namedtuple('Response', 'label error_message')
    if trunc and len(label) > length:
        label = label[0: length]
    if clip and len(label) > length:
        label = label[len(label) - length: len(label)]

    error_message = None
    if len(label) > length:
        error_message = ("The specified label, '%s', has too many characters."
                         " Please enter a label of %s or fewer "
                         "characters.") % (label, length)

    return r(label=label, error_message=error_message)


def recursive_search(params, class_name=None, method=None, args={}):
    """Recursive params checker

    Iterates through all items of a passed in dict, tuple, or list
    and either returns everything as a list, returns an optional subset of them
    or calls a specified method on the subset

    Example Usage:

    .. code-block:: python

        from autoprotocol.protocol import Protocol
        from autoprotocol.container import Well
        from autoprotocol_utilities import recursive_search

        p = Protocol()

        example_container = p.ref(name="example", id=None, cont_type="96-pcr",
                                  storage="ambient")

        p.dispense(ref=example_container, reagent="water",
                   columns=[{"column": 0, "volume": "20:microliters"},
                            {"column": 1, "volume": "10:microliters"}
                            ])

        wells = example_container.wells_from(start=0, num=9, columnwise=True)

        recursive_search(params=wells.wells, class_name=Well,
                         method=volume_check, args={"usage_volume": 15})
    Returns:

    .. code-block:: python

        ["1 volume errors: You want to pipette 15.0 ul from a container with "
        "3.0 ul dead volume (18.0 ul total). However, your aliquot: example-1,"
        "only has 10.0 ul."]

    Parameters
    ----------
    params : list, tuple or dict
        Structure to parse
    class_name : Class name, optional
        Optionally return only instances of a class.
    method : function, optional
        A function that will be applied to all instances found of a class,
        must include class name.
    args : parameters, optional
        Parameters to pass to a method, if desired.

    Returns
    -------
    list
        Will return a list of all items, or the found items of a specified
        class, or the response (if not None) from a method called on found
        items.


    """

    all_fields = []

    def find_all_fields(params):
        if isinstance(params, dict):
            for key, value in params.items():
                all_fields.append(key)
                find_all_fields(value)
        elif isinstance(params, list) or isinstance(params, tuple):
            for item in params:
                find_all_fields(item)
        else:
            all_fields.append(params)

    find_all_fields(params)

    if class_name:
        found_instances = []
        for field in all_fields:
            if isinstance(field, class_name):
                found_instances.append(field)
        if method:
            method_msgs = []
            if hasattr(method, '__call__'):
                for found in found_instances:
                    response = method(found, **args)
                    if response is not None:
                        method_msgs.append(response)
            else:
                raise Exception("Method called has no method __call__")
            return method_msgs
        else:
            return found_instances
    else:
        return all_fields


def transfer_properties(src_wells, dest_wells, properties={}, args={},
                        pset=False):
    """Transfer all or select propeties from one well to another

    Uses add_properties to transfer the properties from one well to another

    Parameters
    ----------
    src_wells: well, list of wells, WellGroup
    dest_wells: well, list of wells, WellGroup
    properties: dict, optional
        Dict with properties to transfer as keys and a function to modify
        property as value. If no modification is required, put None.
        If the dict is empty all propeties will be transferred.
    args: dict, optional
        Dict of dicts where the key is the property used to indicate the
        function and the value another dict containing the arguments.
    pset: bool, optional
        Indicate where to set or add the property, defaults to add.

    Returns
    -------
    list
        List of strings if some properties could not be found
    None
        If no messages were found

    Raises
    ------
    ValueError
        If src_wells or dest_wells are not of type well
    ValueError
        If src_wells and dest_wells are not of equal length
    ValueError
        If properties is not of type dict
    """

    if isinstance(src_wells, Well):
        src_wells = [src_wells]
    if isinstance(dest_wells, Well):
        dest_wells = [dest_wells]

    assert isinstance(src_wells, (list, WellGroup))
    assert isinstance(dest_wells, (list, WellGroup))
    assert len(src_wells) == len(dest_wells)
    assert isinstance(properties, dict)
    for well in list(src_wells) + list(dest_wells):
        assert isinstance(well, Well)

    def dummy(prop):
        return prop

    error_messages = []

    if len(properties) == 0:
        for i, well in enumerate(src_wells):
            if pset:
                dest_wells[i].set_properties(well.properties)
            else:
                dest_wells[i].add_properties(well.properties)
    else:
        for prop, func in properties.items():
            if func is None:
                func = dummy
            for i, well in enumerate(src_wells):
                if prop in well.properties:
                    if pset:
                        dest_wells[i].set_properties(
                            {prop: func(well.properties[prop],
                                        **args.get(prop, {}))})
                    else:
                        dest_wells[i].add_properties(
                            {prop: func(well.properties[prop],
                                        **args.get(prop, {}))})
                else:
                    error_messages.append("Could not find property %s on "
                                          "well%s." % (prop, well))

    if len(error_messages) > 0:
        return error_messages
    else:
        return None



from autoprotocol.container import Container, WellGroup, Well
from autoprotocol.container_type import _CONTAINER_TYPES
from autoprotocol.unit import Unit
from .misc_helpers import flatten_list
from .rectangle import binary_list, chop_list, max_rectangle, \
    get_quadrant_binary_list, get_well_in_quadrant
from collections import namedtuple, Counter
from operator import itemgetter
import math
import sys

if sys.version_info[0] >= 3:
    string_type = str
else:
    string_type = basestring


def list_of_filled_wells(wells, empty=False):
    """
    For the container given, determine which wells are filled

    Parameters
    ----------
    wells : Container, WellGroup, list
        Takes a container (uses all wells), a WellGroup or a List of wells
    empty : bool
        If True return empty wells instead of filled

    Returns
    -------
    list
        list of wells

    Raises
    ------
    ValueError
        If wells are not of type list, WellGroup or Container

    """
    assert isinstance(wells, (Container, WellGroup, list))
    if isinstance(wells, Container):
        wells = wells.all_wells()

    return_wells = []
    for well in wells:
        if not empty:
            if well.volume is not None:
                return_wells.append(well)
        if empty:
            if well.volume is None:
                return_wells.append(well)
    return return_wells


def first_empty_well(wells, return_index=True):
    """
    Get the first empty well of a container followed by only empty wells

    Parameters
    ----------
    wells : Container, WellGroup, list
        Can accept a container, WellGroup or list of wells.
    return_index : bool, optional
        Default true, if true returns the index of the well, if false the
        well itself.

    Returns
    -------
    well
        The first empty well OR
    int
        The index of the first empty well when return_index=True OR
    None
        when no empty well was found.

    Raises
    ------
    ValueError
        If wells are not of type list, WellGroup or Container

    """
    assert isinstance(wells, (Container, WellGroup, list))
    if isinstance(wells, Container):
        wells = list(wells.all_wells())
    else:
        assert len(unique_containers(wells)) == 1
        wells = list(sort_well_group(wells))

    last_well = max(wells, key=lambda x: x.index if x.volume else 0)
    next_index = wells.index(last_well) + 1
    if len(wells) > next_index:
        well = wells[next_index]
    else:
        well = None

    if return_index and well:
        well = well.index

    return well


def unique_containers(wells):
    """Get unique containers

    Get a list of unique containers for a list of wells

    Example Usage:

    .. code-block:: python

        from autoprotocol import Protocol
        from autoprotocol_utilities import get_well_list_by_cont

        p = Protocol()

        wells_1 = p.ref("plate_1_96", None, "96-flat",
                        discard=True).wells_from("A1", 12)
        wells_2 = p.ref("plate_2_96", None, "96-flat",
                        discard=True).wells(0, 24, 49)
        wells_3 = p.ref("plate_3_96", None, "96-flat",
                        discard=True).well("D3")

        many_wells = wells_1 + wells_2 + wells_3
        unique_containers(many_wells)

    Returns:

    .. code-block:: python

        [Container(plate_1_96), Container(plate_3_96), Container(plate_2_96)]

    Parameters
    ----------
    wells : Well, list, WellGroup
        List of wells.

    Returns
    -------
    list
        List of Containers

    Raises
    ------
    ValueError
        If wells are not of type list or WellGroup

    """
    if isinstance(wells, Well):
        wells = [wells]
    assert isinstance(wells, (list, WellGroup)), "unique_containers requires"
    " a Well, list of wells or a WellGroup"
    if isinstance(wells, WellGroup):
        wells = list(wells)
    wells = flatten_list(wells)
    cont = list(set([well.container for well in wells]))
    return cont


def sort_well_group(wells, columnwise=False):
    """Sort a well group in rowwise or columnwise format.

    This function sorts first by container id and name, then by row or
    column, as needed. This function is useful to sort a list of wells
    passed in an unknown order (eg. user entered).

    Parameters
    ----------
    wells : list, WellGroup
        List of wells to sort.
    columnwise : bool, optional

    Returns
    -------
    WellGroup
        Sorted list of wells

    Raises
    ------
    ValueError
        If wells are not of type list or WellGroup
    ValueError
        If elements of wells are not of type well

    """
    if isinstance(wells, list):
        for well in wells:
            assert isinstance(well, Well)
        wells = WellGroup(wells)
    assert isinstance(wells, WellGroup), "wells must be an instance"
    " of the WellGroup class or of type list"
    well_list = [(
        well,
        well.container.id,
        well.container.name,
        well.container.decompose(well)[0],
        well.container.decompose(well)[1]
    ) for well in wells
    ]

    if columnwise:
        sorted_well_list = sorted(well_list, key=itemgetter(1, 2, 4, 3))
    else:
        sorted_well_list = sorted(well_list, key=itemgetter(1, 2, 3, 4))

    sorted_well_group = WellGroup([well[0] for well in sorted_well_list])
    return sorted_well_group


def stamp_shape(wells, full=True, quad=False):
    """Determine if a list of wells is stampable

    Find biggest reactangle that can be stamped from a list of wells. Can be
    any rectangle, or enforce full row or column span.
    If a list of wells from a container that cannot be stamped is provided,
    all wells will be returned in `remaining_wells` of the stamp shape.

    Example Usage:

    .. code-block:: python

        from autoprotocol import Protocol
        from autoprotocol_utilities import stamp_shape, first_empty_well, \
            flatten_list

        p = Protocol()
        plate = p.ref("myplate", cont_type="96-pcr", storage="cold_4")
        dest_plate = p.ref("newplate", cont_type="96-pcr", storage="cold_4")
        src_wells = plate.wells_from(0, 40)
        shape = stamp_shape(src_wells)
        remaining_wells = []
        for s in shape:
            p.stamp(s.start_well, dest_plate.well(0),
                    "10:microliter", s.shape)
            remaining_wells.append(s.remaining_wells)
        next_dest = first_empty_well(dest_plate)
        remaining_wells = flatten_list(remaining_wells)
        p.transfer(remaining_wells,
                   dest_plate.wells_from(next_dest, len(remaining_wells)),
                   "10:microliter"
                   )

    Autoprotocol Output:

    .. code-block:: json

        {
            "refs": {
                "myplate": {
                    "new": "96-pcr",
                    "store": {
                        "where": "cold_4"
                    }
                },
                "newplate": {
                    "new": "96-pcr",
                    "store": {
                        "where": "cold_4"
                    }
                }
            },
            "instructions": [
                {
                    "groups": [
                        {
                            "transfer": [
                                {
                                    "volume": "10.0:microliter",
                                    "to": "newplate/0",
                                    "from": "myplate/0"
                                }
                            ],
                            "shape": {
                                "rows": 3,
                                "columns": 12
                            },
                            "tip_layout": 96
                        }
                    ],
                    "op": "stamp"
                },
                {
                    "groups": [
                        {
                            "transfer": [
                                {
                                    "volume": "10.0:microliter",
                                    "to": "newplate/36",
                                    "from": "myplate/36"
                                }
                            ]
                        },
                        {
                            "transfer": [
                                {
                                    "volume": "10.0:microliter",
                                    "to": "newplate/37",
                                    "from": "myplate/37"
                                }
                            ]
                        },
                        {
                            "transfer": [
                                {
                                    "volume": "10.0:microliter",
                                    "to": "newplate/38",
                                    "from": "myplate/38"
                                }
                            ]
                        },
                        {
                            "transfer": [
                                {
                                    "volume": "10.0:microliter",
                                    "to": "newplate/39",
                                    "from": "myplate/39"
                                }
                            ]
                        }
                    ],
                    "op": "pipette"
                }
            ]
        }

    Parameters
    ----------
    wells: Container, WellGroup, list
        If Container - all filled wells will be used to determine the shape.
        If list of wells or well_group all provided wells will be analyzed.
    full: bool, optional
        If true will only return shapes that span either the full rows or
        columns of the container.
    quad: bool, optional
        Set to true if you want to get the stamp shape for a 384 well testing
        all quadrants. False is used for determining col- vs row-wise. True
        is used to initiate the correct stamping.

    Returns
    -------
    list
        contains namedtuples where each tuple has the following parameters

    start_well: well
        is the top left well for the source stamp group
    shape: dict
        is a dict of `rows` and `columns` describing the stamp shape
    remainging_wells: list
        is a list of wells that are not included in the stamp shape
    included_wells: list
        is a list of wells that is included in the stamp shape


    Raises
    ------
    RuntimeError
        If wells are not of type list or WellGroup
    ValueError
        If elements of wells are not of type well
    ValueError
        If wells are not from one container only

    """
    if isinstance(wells, Container):
        cont = wells
        wells = list_of_filled_wells(wells)
    elif isinstance(wells, (list, WellGroup)):
        assert len(unique_containers(wells)) == 1, "Stamp_shape: wells have "
        "to come from one container"
        for well in wells:
            assert isinstance(well, Well), "Stamp_shape: elements of wells"
            " have to be of type Well"
        wells = sort_well_group(wells)
        cont = unique_containers(wells)[0]
    else:
        raise RuntimeError("Stamp_shape: wells has to be a list or a "
                           "WellGroup")

    def make_stamp_tuple(r, rows, cols, q=None):
        height = r.height
        width = r.width
        if full:
            if not (r.height == rows or r.width == cols):
                height = 0
                width = 0
        if width != 0 or height != 0:
            start_index = (r.y * cols) + r.x
        else:
            start_index = None
        wells_included = []
        for y in range(height):
            for z in range(width):
                wells_included.append(start_index + y * cols + z)
        if q is not None:
            wells_included = get_well_in_quadrant(wells_included, q)
            if width != 0 or height != 0:
                start_index = get_well_in_quadrant([start_index], q)[0]

        wells_idx_remaining = [x.index for x in wells if x.index not in
                               wells_included]
        if start_index is not None:
            start_well = [x for x in wells if x.index == start_index][0]
        else:
            start_well = start_index
        wells_remaining = [x for x in wells if x.index in wells_idx_remaining]
        wells_included = [x for x in wells if x.index in wells_included]
        r = stamp_shape(start_well=start_well,
                        shape=dict(rows=height, columns=width),
                        remaining_wells=wells_remaining,
                        included_wells=wells_included)
        return r

    stamp_shape = namedtuple(
        'Stamp', 'start_well shape remaining_wells included_wells')
    rows = cont.container_type.row_count()
    cols = cont.container_type.col_count
    well_count = cont.container_type.well_count
    indices = [x.index for x in wells]

    if well_count not in (96, 384):
        shape = stamp_shape(start_well=None,
                            shape=dict(rows=0, columns=0),
                            remaining_wells=wells,
                            included_wells=[])
        return [shape]

    bnry_list = [bnry for bnry in binary_list(indices, length=well_count)]
    if well_count == 384 and quad:
        bnry_list_list = get_quadrant_binary_list(bnry_list)
        temp_shape = []
        temp_remaining_wells = []
        remaining_wells = []
        for i, bnry_list in enumerate(bnry_list_list):
            bnry_mat = chop_list(bnry_list, 12)
            r = max_rectangle(bnry_mat, value=1)
            temp_shape.append(make_stamp_tuple(r, rows / 2, cols / 2, i))
            temp_remaining_wells.append(temp_shape[i].remaining_wells)
        temp_remaining_wells = Counter(flatten_list(temp_remaining_wells))
        for k, v in temp_remaining_wells.items():
            if v == 4:
                remaining_wells.append(k)
        shape = []
        for s in temp_shape:
            shape.append(stamp_shape(start_well=s.start_well,
                                     shape=s.shape,
                                     remaining_wells=remaining_wells,
                                     included_wells=s.included_wells))
    else:
        bnry_mat = chop_list(bnry_list, cols)
        r = max_rectangle(bnry_mat, value=1)
        shape = [make_stamp_tuple(r, rows, cols)]

    return shape


def is_columnwise(wells):
    """Detect if input wells are in a columnwise format.

    This function only triggers if the first column is full and only a
    consecutive fractionally filled columns exist. It is used to determine
    whether `columnwise` should be used in a pipette operation.

    Only accepts wells that belong to 1 container. Use unique_containers or
    get_well_list_by_cont to assure you submit only wells from one container.

    Patterns detected (4x6 plate):

    .. code-block:: none

        x x   | x x x  | x | x
        x     | x x x  | x | x
        x     | x x x  | x | x
        x     | x x x  | x |

    Patterns NOT detected (4x6 plate):

    .. code-block:: none

        x x x  | x      |
          x x  | x      |
          x x  | x      |
          x x  | x x x  |

    Example Usage:

    .. code-block:: python

        from autoprotocol.protocol import Protocol
        from autoprotocol_utilities import is_columnwise

        p = Protocol()
        plate = p.ref("plate", None, cont_type="96-flat", storage="cold_4",
                      cover="standard")
        col_wells = plate.wells_from(start="A1", num=17, columnwise=True)
        col_wells_2 = plate.wells_from(start="A2", num=17, columnwise=True)
        row_wells = plate.wells_from(start="A1", num=17, columnwise=False)
        rand_wells = plate.wells("A2", "B12", "H7")

        is_columnwise(col_wells)
        is_columnwise(col_wells_2)
        is_columnwise(row_wells)
        is_columnwise(rand_wells)

    Returns:

    .. code-block:: python

        True
        True
        False
        False

    Parameters
    ----------
    wells: Well, list, WellGroup
        List of wells or well_group containing the wells in question.

    Returns
    -------
    bool
        True if columnwise. False if rowwise.
    list
        List of strings if errors were encountered.

    Raises
    ------
    ValueError
        If wells are not of type Well, list or WellGroup
    ValueError
        If elements of wells are not of type Well

    """

    em = []
    colwise = False
    if isinstance(wells, Well):
        colwise = False
    else:
        assert isinstance(wells, (list, WellGroup)), "is_columnwise: wells"
        " has to be a list or a WellGroup"
        for well in wells:
            assert isinstance(well, (Well)), "is_columnwise: elements of "
            "wells have to be of type Well"
        if len(unique_containers(wells)) != 1:
            em.append("is_columnwise: wells have to come from one container")

    cont = unique_containers(wells)[0]

    all_wells = list(cont.all_wells(columnwise=True))
    top_wells = list(cont.wells_from(0, cont.container_type.col_count))
    wells = sort_well_group(wells, columnwise=True)

    if wells[0] in top_wells:
        top_well = top_wells[top_wells.index(wells[0])]
        start = all_wells.index(top_well)
        for x in range(len(wells)):
            if wells[x] == all_wells[start]:
                colwise = True
                start += 1
            else:
                colwise = False
                break
    else:
        colwise = False

    em = [_f for _f in em if _f]
    if len(em) > 0:
        return em
    else:
        return colwise


def plates_needed(wells_needed, wells_available):
    """
    Takes wells needed as a numbers (int or float)
    and wells_available as a container, container type string or a well number
    (int or float) and calculates how many plates are
    needed to accomodate the wells_needed.

    Parameters
    ----------
    wells_needed: float, int
        How many you need
    wells_available: Container, float, int, string
        How many you have available per unit. If container or a string
        identifying a container type, then all wells of
        this container will be considered

    Returns
    -------
    int
        How many of unit you will need to accomodate all wells_needed

    Raises
    ------
    RuntimeError
        If wells_needed is not of type integer or float
    RuntimeError
        If wells_available is not of type integer or float or Container

    """
    if not isinstance(wells_needed, (float, int)):
        raise RuntimeError("wells_needed has to be an int or a float")
    if isinstance(wells_available, Container):
        wells_available = float(wells_available.container_type.well_count)
    elif isinstance(wells_available, string_type):
        cont = _CONTAINER_TYPES.get(wells_available, None)
        if cont:
            wells_available = float(cont.well_count)
        else:
            raise RuntimeError("If `wells_available` is a string, it has to "
                               "match a valid `container_type`. %s does not "
                               "match any container type" % wells_available)
    elif isinstance(wells_available, int):
        wells_available = float(wells_available)
    elif not isinstance(wells_available, (float, int, Container, string_type)):
        raise RuntimeError("wells_available has to be a container, a string "
                           "uniquely identifying a container type, "
                           "int or float")
    return int(math.ceil(wells_needed / wells_available))


def set_pipettable_volume(well, use_safe_vol=False):
    """Remove dead volume from pipettable volume.

    In one_tip true pipetting operations the volume of the well is used to
    determine who many more wells can be filled from this source well. Thus
    it is useful to remove the dead_volume (default), or the safe minimum from
    the set_volume of the well.
    It is recommeneded to remove the dead_volume only and check for safe_vol
    later.

    Parameters
    ----------
    well : Container, WellGroup, list, Well
        Well to set.
    use_safe_vol : bool, optional
        Instead of removing the indicated dead_volume, remove the safe minimum
        volume.

    Returns
    -------
    Container, WellGroup, list, Well
        Will return the same type as was received

    """

    cont = {}
    if isinstance(well, (list, WellGroup)):
        cont = get_well_list_by_cont(well)
    elif isinstance(well, Container):
        cont[well] = list_of_filled_wells(well)
    elif isinstance(well, Well):
        cont[well.container] = [well]

    for c, w in cont.items():
        correction_vol = c.container_type.dead_volume_ul
        if use_safe_vol:
            correction_vol = c.container_type.safe_min_volume_ul
        for x in w:
            x.set_volume(x.volume - correction_vol)

    return well


def volume_check(well, usage_volume=0, use_safe_vol=False,
                 use_safe_dead_diff=False):
    """Basic Volume check

    Checks to see if the designated well has usage_volume above the well's
    dead volume. In other words, this method checks if usage_volume can be
    pipetted out of well.

    Example Usage:

    .. code-block:: python

            from autoprotocol import Protocol
            from autoprotocol_utilities.container_helpers import volume_check

            p = Protocol()
            example_container = p.ref(name="exampleplate", id=None,
                                      cont_type="96-pcr", storage="warm_37")
            p.dispense(ref=example_container, reagent="water",
                       columns=[{"column": 0, "volume": "10:microliters"}])

            #Checks if there are 5 microliters above the dead volume
            #available in well 0
            assert (volume_check(well=example_container.well(0),
                    usage_volume=5)) is None
            #Checks if the volume in well 0 is at least the safe minimum volume
            assert (volume_check(well=example_container.well(0),
                    usage_volume=0, use_safe_vol=True) is None

    Parameters
    ----------
    well : Well, WellGroup, list
        Well(s) to test
    usage_volume : Unit, str, int, float, optional
        Volume to test for. If 0 the aliquot will be tested against the
        container dead volume. If int or float is used, microliter will be
        assumed.
    use_safe_vol : bool, optional
        Use safe minimum volume instead of dead volume
    use_safe_dead_diff : bool, optional
        Use the safe_minimum_volume - dead_volume as the required amount.
        Useful if `set_pipettable_volume()` was used before to correct the
        well_volume to not include the dead_volume anymore

    Returns
    -------
    str
        string of errors if volume check failed OR
    None
        If no errors are detected

    Raises
    ------
    ValueError
        If well is not of type Well, list or WellGroup
    ValueError
        If elements of well are not of type Well

    """

    assert isinstance(well, (Well, WellGroup, list))
    if isinstance(well, Well):
        well = [well]

    error_message = []
    # noinspection PyTypeChecker
    for aliquot in well:
        assert isinstance(aliquot, Well)
        if isinstance(usage_volume, (int, float)):
            usage_volume = Unit(usage_volume, "microliter")
        if isinstance(usage_volume, string_type):
            usage_volume = int(usage_volume.split(":microliter")[0])
        if not aliquot.volume:
            error_message.append(
                "Your aliquot does not have a volume. (%s) We assume 0 uL "
                "for this test." % aliquot)

        correction_vol = aliquot.container.container_type.dead_volume_ul
        message_string = "dead volume"
        volume = Unit(0, "microliter")
        if aliquot.volume:
            volume = aliquot.volume
        if use_safe_vol:
            correction_vol = \
                aliquot.container.container_type.safe_min_volume_ul
            message_string = "safe minimum volume"
        elif use_safe_dead_diff:
            correction_vol = \
                aliquot.container.container_type.safe_min_volume_ul - \
                aliquot.container.container_type.dead_volume_ul
            message_string = "safe minimum volume"
            volume = volume + aliquot.container.container_type.dead_volume_ul
        test_vol = correction_vol + usage_volume

        if test_vol > volume:
            if usage_volume == 0:
                error_message.append(
                    "You want to pipette from a container with {:~P} {!s}. "
                    "However, your aliquot: {!s}, only has {:~P}.".format(
                        correction_vol, message_string,
                        well_name(aliquot), volume))
            else:
                error_message.append(
                    "You want to pipette {:~P} from a container with {:~P} "
                    "{!s} ({:~P} total). However, your aliquot: {!s}, only has"
                    " {:~P}.".format(
                        usage_volume, correction_vol, message_string,
                        usage_volume + correction_vol,
                        well_name(aliquot), volume))
    if error_message:
        error_message = str(len(error_message)) + " volume errors: " + \
            ", ".join(error_message)
    else:
        error_message = None
    return error_message


def well_name(well, alternate_name=None, humanize=False):
    """Determine new well name

    Determine the name for a new well based on passed well.

    Parameters
    ----------
    well: Well
        Well to source original name, index properties.
    alternate_name: str, optional
        If this parameter is passed and the well does not have a name, this
        name will be returned instead of the container name, appended with
        the well index.

    Returns
    -------
    str
        well name in the format `name` if the well had a name or `name-index`
        if the name is derived from the container name or alternate_name

    Raises
    ------
    ValueError
        If well is not of type Well
    ValueError
        It alternate_name is not of type string
    """

    assert isinstance(well, Well)
    if alternate_name:
        assert isinstance(alternate_name, string_type)

    if humanize:
        well_index = well.container.humanize(well.index)
    else:
        well_index = well.index

    if well.name is not None:
        base_name = well.name
    elif alternate_name:
        base_name = "%s-%s" % (alternate_name, well_index)
    else:
        base_name = "%s-%s" % (well.container.name, well_index)

    return base_name


def container_type_checker(containers, shortname, exclude=False):
    """Verify container is of specified container_type.

    Parameters
    ----------
    containers : Container, list
        Single Container or list of Containers
    shortname : str, list of str
        Short name used to specify ContainerType.
    exclude: bool, optional
        Verify container is NOT of specified container_type.
    Returns
    -------
    str
        String of containers failing container_type_check OR
    None
        If no container fails

    Raises
    ------
    ValueError
        If an unknown ContainerType shortname is passed.
    ValueError
        If an containers are not of type Container.
    """
    if isinstance(shortname, list):
        for short in shortname:
            assert short in _CONTAINER_TYPES, ("container_type_check: unknown"
                                               " container shortname: %s , "
                                               "(known types: %s)" %
                                               (short,
                                                str(list(
                                                    _CONTAINER_TYPES.keys()))))
    elif isinstance(shortname, str):
        assert shortname in _CONTAINER_TYPES, ("container_type_check: unknown"
                                               " container shortname: %s , "
                                               "(known types: %s)" %
                                               (shortname,
                                                str(list(
                                                    _CONTAINER_TYPES.keys()))))
    if isinstance(containers, list):
        for cont in containers:
            assert isinstance(cont, Container), ("container_type_check: "
                                                 "containers to check"
                                                 "must be of type Container")
    elif isinstance(containers, Container):
        containers = [containers]
    else:
        raise ValueError(
            "container_type_check: containers to check must be of type "
            "Container")

    error_containers = []
    error_message = None

    for cont in containers:
        if exclude:
            if cont.container_type.shortname in shortname:
                error_containers.append(str(cont))
        else:
            if cont.container_type.shortname not in shortname:
                error_containers.append(str(cont))

    if error_containers:
        message_ending = ' not of the required type(s): ' + \
                         ', '.join(shortname)
        if exclude:
            message_ending = ' of the excluded type(s): ' + \
                             ', '.join(shortname)
        error_message = "Incompatible container(s) found : " + \
                        ', '.join(error_containers) + message_ending

    return error_message


def get_well_list_by_cont(wells):
    """Get wells sorted by container

    Example Usage:

    .. code-block:: python

        from autoprotocol import Protocol
        from autoprotocol_utilities import get_well_list_by_cont

        p = Protocol()

        wells_1 = p.ref("plate_1_96", None, "96-flat",
                        discard=True).wells_from("A1", 12)
        wells_2 = p.ref("plate_2_96", None, "96-flat",
                        discard=True).wells(0, 24, 49)
        wells_3 = p.ref("plate_3_96", None, "96-flat",
                        discard=True).well("D3")

        many_wells = wells_1 + wells_2 + wells_3
        get_well_list_by_cont(many_wells)

    Returns:

    .. code-block:: python

        {
            Container(plate_1_96): [
                Well(Container(plate_1_96), 0, None),
                Well(Container(plate_1_96), 1, None),
                Well(Container(plate_1_96), 2, None)
                ],
            Container(plate_3_96): [
                Well(Container(plate_3_96), 38, None)
                ],
            Container(plate_2_96): [
                Well(Container(plate_2_96), 0, None),
                Well(Container(plate_2_96), 24, None),
                Well(Container(plate_2_96), 49, None)
                ]
        }

    Parameters
    ----------
    wells: list, WellGroup
        The list of wells to be sorted by the containers that they are in

    Returns
    -------
    dict
        Dict with containers as keys and List of wells as value

    Raises
    ------
    ValueError
        If wells is not of type list or WellGroup
    ValueError
        If elements of wells are not of type Well

    """
    assert isinstance(wells, (list, WellGroup))
    for well in wells:
        assert isinstance(well, Well)

    conts = unique_containers(wells)
    well_map = {}
    for cont in conts:
        well_map[cont] = []
        well_map[cont].extend(
            [well for well in wells if well.container == cont])

    return well_map


def next_wells(target, num=1, columnwise=False):
    '''
    Given a plate, a list of plates or a WellGroup, returns a generator
    function that can be used to iterate through the (container's) wells.

    Example Usage:

    .. code-block:: python

        from autoprotocol import Protocol
        from autoprotocol_utilities import next_wells

        p = Protocol()
        c1 = p.ref("c1", id=None, cont_type="96-pcr", discard=True)

        assay_wells = next_wells(c1, num=2, columnwise=False)
        my_wells = []
        your_wells = []
        my_wells.extend(next(assay_wells))
        your_wells.extend(next(assay_wells))
        your_wells.extend(next(assay_wells))
        your_wells

    Returns:

    .. code-block:: python

        [
            Well(Container(c1), 2, None),
            Well(Container(c1), 3, None),
            Well(Container(c1), 4, None),
            Well(Container(c1), 5, None)
        ]

    Parameters
    ----------

    target: Container, List of Containers, WellGroup
        The generator will iteratively return wells from these plate(s)
        or WellGroup, in order, from the first well in the list,
        based on the parameters below.
    num: int, optional
        The generator will produce this many wells with each iteration.
        Defaults to 1.
    columnwise: bool, optional
        Set to True if wells should be generated in columnwise
        format. Defaults to False. If a WellGroup is given this parameter is
        ignored.

    Returns
    -------

    generator:
        This function will iteratively generate the next set of
        wells from the plate. Get the next plates using
        `next(generator)`. Wells will be returned as a
        WellGroup.

    Raises
    ------
    ValueError
        If `target` is not a Container, a list of Containers or a WellGroup
    StopIteration
        If all wells have been used
    '''
    next_index = 0
    assert isinstance(target, (list, Container, WellGroup)), (
        "target must be a Container or a list of Containers or WellGroup")

    well_list = []
    if isinstance(target, Container):
        well_list.extend(target.all_wells(columnwise=columnwise))
    if isinstance(target, list):
        for p in target:
            assert isinstance(p, Container), ("all elements of `target` "
                                              "must be Containers")
        for t in target:
            well_list.extend(t.all_wells(columnwise=columnwise))
    elif isinstance(target, WellGroup):
        well_list = list(target)
    while next_index <= len(well_list) - num:
        yield well_list[next_index:(next_index + num)]
        next_index += num



from .container_helpers import volume_check, set_pipettable_volume, plates_needed, sort_well_group, unique_containers, is_columnwise, stamp_shape, first_empty_well, list_of_filled_wells, well_name, container_type_checker, get_well_list_by_cont, next_wells  # NOQA
from .misc_helpers import user_errors_group, char_limit, printdatetime, printdate, make_list, flatten_list, det_new_group, recursive_search, transfer_properties  # NOQA
from .resource_helpers import ResourceIDs, oligo_scale_default, return_dispense_media, return_agar_plates, ref_kit_container, oligo_dilution_table  # NOQA
from .thermocycle_helpers import melt_curve, thermocycle_ramp  # NOQA
from .bio_calculators import dna_mass_to_mole, dna_mole_to_mass, molar_to_mass_conc, mass_conc_to_molar, ligation_insert_ng, ligation_insert_volume, ligation_insert_amount  # NOQA



from autoprotocol_utilities import list_of_filled_wells
from autoprotocol.container import Container
import sys

if sys.version_info[0] >= 3:
    string_type = str
else:
    string_type = basestring


def get_mag_amplicenter(plate, amplitude_fraction=1.0):
    """Determine amplitude and center for KF operations

    Given the container this function determines the well type and fill volume.
    From that it calculates the center for the magnetic operation (1/2 the
    fill volume) and the amplitude to move around that from 0 to center +
    amplitude.

    Example Usage:

    .. code-block:: python

        from autoprotocol import Protocol
        from autoprotocol_utilities.magnetic_helpers import get_mag_amplicenter

        p = Protocol()
        example_plate = p.ref(name="Example", id=None, cont_type="96-pcr",
                              storage="ambient")
        p.dispense(ref=example_plate, reagent="water",
                   columns=[{"column": 4, "volume": "100:microliters"},])

        get_mag_amplicenter(example_plate)

    Returns:

    .. code-block:: json

        {'center': 0.3125, 'amplitude': 0.3125}

    Parameters
    ----------
    plate: Container
        Plate that is being used
    amplitude_fraction: float, optional
        By default the full available amplitude will be used (from center to
        bottom of well). Use this parameter to reduce the amplitude.

    Returns
    -------
    nametuple
        `center` and `amplitude` both of type float

    Raises
    ------
    ValueError
        If plate is not of type `Container`
    ValueError
        If `amplitude_fraction` is not a float or bigger than 1
    """

    assert isinstance(plate, Container)
    assert isinstance(amplitude_fraction, float)
    assert amplitude_fraction <= 1.0
    max_cont_vol = plate.container_type.well_volume_ul
    max_vol = max([x.volume for x in list_of_filled_wells(plate)])

    ratio = float(max_vol.to_base_units() / max_cont_vol.to_base_units())

    return {"center": ratio / 2, "amplitude": ratio / 2 / amplitude_fraction}


def get_mag_frequency(plate, speed):
    """Determine frequency for KF operation

    Based on plate type and the desired KF speed, determine the frequency for
    the KF operation. Needed for `mix` and `release` operations.

    Example Usage:

    .. code-block:: python

        from autoprotocol import Protocol
        from autoprotocol_utilities.magnetic_helpers import get_mag_frequency

        p = Protocol()
        example_plate = p.ref(name="Example", id=None, cont_type="96-pcr",
                              storage="ambient")
        get_mag_frequency(plate=example_plate, speed="slow")

    Returns:

    .. code-block:: none

        0.4:hertz

    Parameters
    ----------
    plate: Container
        Container that is being used
    speed: string
        String defining the speed - can be `slow`, `medium`, `fast`

    Returns
    -------
    string
        Frequency in the form of "2.5:Hz"

    Raises
    ------
    ValueError
        If `plate` is not of type `Container`
    ValueError
        If `speed` is not 'slow', 'medium', 'fast'
    ValueError
        If plate type is not a key in `frequencies`
    """
    name = plate.container_type.shortname

    frequencies = {"96-deep-kf": {"slow": "0.15:hertz", "medium": "1.5:hertz",
                                  "fast": "2.5:hertz"},
                   "96-deep": {"slow": "0.15:hertz", "medium": "1.5:hertz",
                               "fast": "2.5:hertz"},
                   "96-v-kf": {"slow": "0.5:hertz", "medium": "4.5:hertz",
                               "fast": "9.5:hertz"},
                   "96-flat": {"slow": "0.5:hertz", "medium": "5.5:hertz",
                               "fast": "11.5:hertz"},
                   "96-pcr": {"slow": "0.4:hertz", "medium": "4:hertz",
                              "fast": "8.5:hertz"}}
    assert isinstance(plate, Container)
    assert speed in ['slow', 'medium', 'fast']
    assert name in frequencies.keys()

    return frequencies[name][speed]



from collections import namedtuple

try:
    reduce = reduce
except NameError:
    from functools import reduce  # py3k  # NOQA

# A column in a histogram
Column = namedtuple('Column', 'height x')
# An area under a histogram
Area = namedtuple('Area', 'width height x')
# A rectangle in a matrix/bitmap
Rect = namedtuple('Rect', 'width height x y')


def area(rect):
    """
    Computes geometrical area for Areas and Rects

    Parameters
    ----------
    rect : Area, Rect, Namedtuple
        Object that needs area calculation done

    Returns
    -------
    int
        The area of the object
    """
    return rect.width * rect.height


def area2rect(area, row_idx):
    """
    Converts an Area (no y coord) into a Rect. The y coord of the
    top left of the Rect is derived from the area height and the row_idx.

    Parameters
    ----------
    area : Area, Rect, Namedtuple
        Object that needs area calculation done
    row_idx: int
        The row index of the area

    Returns
    -------
    Rectangle
        A rectangle object with width, height, x, and y

    """
    return Rect(width=area.width,
                height=area.height,
                x=area.x,
                y=row_idx - area.height + 1)


def max_rectangle(mat, value=0):
    """Find height, width of the largest rectangle containing all `value`'s.
    For each row solve "Largest Rectangle in a Histrogram" problem [1]:
    [1]: http://blog.csdn.net/arbuckle/archive/2006/05/06/710988.aspx

    Parameters
    ----------
    mat: list
        Object that needs area calculation done
    value: int, optional
        Value that user is looking for

    Returns
    -------
    Rectangle
        The maximum sized rectangle in the given mat

    """
    row_idx = 0
    it = iter(mat)
    hist = [int(el == value) for el in next(it, [])]
    max_rect = area2rect(max_histogram_area(hist), row_idx)
    for row in it:
        row_idx += 1
        hist = [(1 + h) if el == value else 0 for h, el in zip(hist, row)]
        max_rect_in_row = area2rect(max_histogram_area(hist), row_idx)
        max_rect = max(max_rect, max_rect_in_row, key=area)
    return max_rect


def max_histogram_area(histogram):
    """Find height, width of the largest rectangle that fits entirely under
    the histogram.
    Algorithm is "Linear search using a stack of incomplete subproblems" [1].
    [1]: http://blog.csdn.net/arbuckle/archive/2006/05/06/710988.aspx

    Parameters
    ----------
    histogram: list
        The histogram that the method will find the largest rectangle within

    Returns
    -------
    Area
        The maximum area of the given histogram

    """
    stack = []
    top = lambda: stack[-1]
    max_area = Area(width=0, height=0, x=0)
    pos = 0  # current position in the histogram
    for pos, height in enumerate(histogram):
        start = pos
        while True:
            if not stack or height > top().height:
                stack.append(Column(height=height, x=start))  # push
            elif stack and height < top().height:
                col_height, col_x = stack.pop()
                stack_area = Area(width=pos - col_x, height=col_height,
                                  x=col_x)
                max_area = max(max_area, stack_area, key=area)
                start = col_x
                continue
            break  # height == top().height goes here

    pos += 1  # equivalent to pos = len(histogram)
    for height, start in stack:
        stack_area = Area(width=pos - start, height=height, x=start)
        max_area = max(max_area, stack_area, key=area)

    return max_area


def binary_list(wells, length=None):
    """Turns a list of indices into a binary list with list at
    indices that appear in the initial list set to 1, and 0
    otherwise. `wells` must be sorted, `length` is the length of
    the resulting list.

    .. code-block:: none

        [bnry for bnry in binary_list([1, 3, 5], length=7)]
        [0, 1, 0, 1, 0, 1, 0]

    Parameters
    ----------
    wells: Sorted list
        The wells that need to be changed into a binary list
    length: Int
        The length of the total list

    Returns
    -------
    None
        Turns wells into a binary list

    """
    length = length or max(wells) + 1
    wells_ptr = 0
    for i in range(length):
        try:
            if wells[wells_ptr] == i:
                wells_ptr += 1
                yield 1
            else:
                yield 0
        except IndexError:
            yield 0


def get_quadrant_indices(quad):
    """Return a list of well indices that correspond to the correct quadrant
    on a 384 well plate

    Parameters
    ----------
    quad: Int
        The wells that need to be changed into a binary list

    Returns
    -------
    List
        All the wells inside the desired quadrant

    """
    assert quad in [0, 1, 2, 3]
    start_well = [0, 1, 24, 25]
    wells = []
    for row_offset in range(start_well[quad], 384, 48):
        for col_offset in range(0, 24, 2):
            wells.append(row_offset + col_offset)

    return wells


def get_quadrant_binary_list(binary_list, quad=[0, 1, 2, 3]):
    """Take a binary list of 384 elements (aka wells) and return all the wells
    in the designated quadrant.
    This will be the stampable 96 wells that we have to check for a rectangle.

    Parameters
    ----------
    binary_list: List
        The 384 element well plate

    quad: list
        The quadrants to look in, the values can be: 0,1,2,3

    Returns
    -------
    List of lists
        A list filled with smaller lists, separating each set of wells into
        quadrants

    """
    assert len(binary_list) == 384
    for q in quad:
        assert q in [0, 1, 2, 3]

    wells = []
    for q in quad:
        wells.append([binary_list[i] for i in get_quadrant_indices(q)])

    return wells


def get_well_in_quadrant(quadwells, quad):
    """Take a list of wells and quadrant and return the correct well index in
    the 384 plate

    Parameters
    ----------
    quadwells: list
        The wells in question
    quad: Int
        The quadrant the well is in

    Returns
    -------
    List
        Returns the correct well indices of all the quad-specific wells in
        quadwells

    """
    assert isinstance(quadwells, list)
    assert quad in [0, 1, 2, 3]

    basewells = get_quadrant_indices(quad)
    wells = [basewells[int(i)] for i in quadwells]

    return wells


def chop_list(lst, chop_length, filler=None):
    """Chops a list into a list of lists. Used to generate a plate map
    corresponding to microplate layouts.

    Parameters
    ----------
    lst: list
        The original list that needs to be chopped up
    chop_length: Int
        The length of the pieces to chop the big list into
    filler: Int, optional
        The stuff to fill up the leftover pieces

    Returns
    -------
    List
        Returns the correct well indices of all the quad-specific wells in
        quadwells

    """
    assert chop_length > 0
    ret = []
    row_idx = 0
    while True:
        # extract next chunk from input list
        row = lst[row_idx * chop_length:(row_idx + 1) * chop_length]
        row_idx += 1
        row_len = len(row)

        # list length was integer multiple of chop_length
        if row_len == 0:
            return ret
        # still in the middle of the list
        elif row_len == chop_length:
            ret.append(row)
        # end of list, but last row needs filling up
        else:
            ret.append(row + [filler] * (chop_length - row_len))
            return ret

    return ret



from autoprotocol.container import Container
from autoprotocol.protocol import Ref
from autoprotocol import Unit
from collections import namedtuple
import sys

if sys.version_info[0] >= 3:
    string_type = str
else:
    string_type = basestring


def oligo_scale_default(length, scale, label):
    """Detects if the oligo length matches the selected scale

    Every oligo provider has a length limit on the oligos based on the
    selected scale. The larger the scale, the longer or shorter the oligo
    can be.
    This function checks for IDT limits

    Parameters
    ----------
    length : int
        Length of the oligo in question.
    scale : str
        Scale of the oligo in question.
    label : str
        Name of the oligo.

    Returns
    -------
    namedtuple
        'success' (bool) and 'error_message' (string) that is empty upon
        success

    """

    r = namedtuple('Response', 'success error_message')
    ok = True
    error_message = None
    scale_ranges = {
        '10nm': [15, 60],
        '25nm': [15, 60],
        '100nm': [10, 90],
        '250nm': [5, 100],
        '1um': [5, 100]
    }

    if scale not in scale_ranges.keys():
        error_message = ("The specified oligo, '{0!s}', does not have a "
                         "recognized scale of {0!s}".format(
                             ', '.join(scale_ranges.keys)))
    else:
        ok = True if ((length >= scale_ranges[scale][0]) & (
            length <= scale_ranges[scale][1])) else False

    if not ok:
        error_message = ("The specified oligo, '{0!s}', is {1!s} base pairs "
                         "long. This sequence length is invalid for the scale"
                         " of synthesis chosen ({2!s}). The acceptable range "
                         "for this scale is {3!s} - {4!s} base pairs "
                         "long".format(label, length, scale,
                                       scale_ranges[scale][0],
                                       scale_ranges[scale][1]))

    return r(success=ok, error_message=error_message)


def oligo_dilution_table(conc=None, sc=None):
    """Return dilution table

    Determine the amount of diluent to add to an oligo based on
    concentration wanted and scale ordered. This function can return the
    entire dilution table or slices and values as needd

    Example Usage:

    .. code-block:: python

        oligo_dilution_table(conc="100uM")
        oligo_dilution_table(conc="100uM", scale="25nm")

    Returns:

    .. code-block:: python

        {"100uM": {"10nm":60, "25nm": 250, "100nm": 1000, "250nm": 2500,
                   "1um": 10000}}
        250

    Parameters
    ----------
    conc : str, optional
        The concentration you wish to select for. Currently 100uM or 1mM.
    sc : str, optional
        The scale you wish to select for. Currently '10nm', '25nm',
        '100nm', '250nm' or '1um'.

    Returns
    -------
    int
        If concentration and scale are given this returns the dilution volume
        only
    dict
        If only one of concentration or scale is provided this function
        returns a dict. If a concentration is selected it returns a dict where
        scales are the keys and dilution volumes the values.
        If a scale is selected it returns a dict with concentrations as keys
        and dicts of scale (key) and dilution volume (value).
        If nothing is selected the full dict of concentrations with dicts of
        scales and dilution volumes is returned.



    Raises
    ------
    ValueError
        If conc is not a valid concentration: '100uM', '1mM'
    ValueError
        If sc is is not a valid scale: '10nm', '25nm', '100nm', '250nm', '1um'

    """
    concentration = '100uM', '1mM'
    scale = '10nm', '25nm', '100nm', '250nm', '1um'
    volumes = [60, 250, 1000, 2500, 10000, 6, 25, 100, 250, 1000]

    if conc:
        assert conc in concentration, ("conc has to be in %s " % concentration)
    if sc:
        assert sc in scale, ("sc has to be in %s " % scale)

    dilution_table = {}
    for i, y in enumerate(concentration):
        dilution_table[y] = dict(
            zip(scale, volumes[i * len(scale):i * len(scale) + len(scale)]))

    if conc and sc:
        return dilution_table[conc][sc]
    elif conc:
        return dilution_table[conc]
    elif sc:
        for x in scale:
            if sc != x:
                for c in dilution_table:
                    dilution_table[c].pop(x, None)
        return dilution_table
    else:
        return dilution_table


def return_agar_plates(wells=6):
    """Returns a dict of all available agar plates.

    Parameters
    ----------
    wells : integer
        Integer of the number of wells in the plate, 1 or 6.
        Default: 6 for 6-well plate

    Returns
    -------
    dict
        plates with plate identity as key and kit_id as value

    Raises
    ------
    ValueError
        If wells is not a integer equal to 1 or 6

    """
    if wells == 6:
        plates = {"lb_miller_50ug_ml_kan": "ki17rs7j799zc2",
                  "lb_miller_100ug_ml_amp": "ki17sbb845ssx9",
                  "lb_miller_100ug_ml_specto": "ki17sbb9r7jf98",
                  "lb_miller_100ug_ml_cm": "ki17urn3gg8tmj",
                  "lb_miller_noAB": "ki17reefwqq3sq"}
    elif wells == 1:
        plates = {"lb_miller_50ug_ml_kan": "ki17t8j7kkzc4g",
                  "lb_miller_100ug_ml_amp": "ki17t8jcebshtr",
                  "lb_miller_100ug_ml_specto": "ki17t8jaa96pw3",
                  "lb_miller_100ug_ml_cm": "ki17urn592xejq",
                  "lb_miller_noAB": "ki17t8jejbea4z"}
    else:
        raise ValueError("Wells has to be an integer, either 1 or 6")
    return (plates)


def return_dispense_media():
    """Returns a dict of media for reagent dispenser.

    Example Usage:

    .. code-block:: python

        from autoprotocol.protocol import Protocol
        from autoprotocol-utilities import return_dispense_media

        p = Protocol()
        media = return_dispense_media()["50_ug/ml_Kanamycin"]
        plate = p.ref("myplate", cont_type="96-flat", discard=True)
        p.dispense_full_plate(plate, media, "200:microliter")

    Autoprotocol Output:
    -------

    .. code-block:: python

        {
            "refs": {
                "myplate": {
                    "new": "96-flat",
                    "discard": True
                }
            },
            "instructions": [
                {
                    "reagent": "lb_miller_50ug_ml_kan",
                    "object": "myplate",
                    "columns": [
                        {
                            "column": 0,
                            "volume": "200:microliter"
                        },
                        {
                            "column": 1,
                            "volume": "200:microliter"
                        },
                        {
                            "column": 2,
                            "volume": "200:microliter"
                        },
                        {
                            "column": 3,
                            "volume": "200:microliter"
                        },
                        {
                            "column": 4,
                            "volume": "200:microliter"
                        },
                        {
                            "column": 5,
                            "volume": "200:microliter"
                        },
                        {
                            "column": 6,
                            "volume": "200:microliter"
                        },
                        {
                            "column": 7,
                            "volume": "200:microliter"
                        },
                        {
                            "column": 8,
                            "volume": "200:microliter"
                        },
                        {
                            "column": 9,
                            "volume": "200:microliter"
                        },
                        {
                            "column": 10,
                            "volume": "200:microliter"
                        },
                        {
                            "column": 11,
                            "volume": "200:microliter"
                        }
                    ],
                    "op": "dispense"
                }
            ]
        }

    Returns
    -------
    dict
        Media with common `display_name` as key and identifier for code
        as value


    """
    media = {"50_ug/ml_Kanamycin": "lb_miller_50ug_ml_kan",
             "100_ug/ml_Ampicillin": "lb_miller_100ug_ml_amp",
             "100_ug/mL_Spectinomycin": "lb_miller_100ug_ml_specto",
             "30_ug/ml_Kanamycin": "lb_miller_30ug_ml_kan",
             "15_ug/ml_Tetracycline": "lb_miller_15ug_ml_tet",
             "50_ug/ml_Kanamycin_25_ug/ml_Chloramphenicol":
             "lb_miller_50ug_ml_kan_25ug_ml_cm",
             "25_ug/ml_Chloramphenicol": "lb_miller_25ug_ml_cm",
             "LB_miller": "lb_miller_noAB",
             "TB_100_ug/ml_Ampicillin": "tb_100ug_ml_amp",
             "TB_50_ug/ml_Kanamycin": "tb_50ug_ml_kan"}
    return (media)


def ref_kit_container(protocol, name, container, kit_id, discard=True,
                      store=None):
    """Reserve agar plates for use within a protocol.

    In use only to allow booking of agar plates within a protocol.

    Example Usage:

    .. code-block:: python

        from autoprotocol import Protocol
        from autoprotocol_utilities import return_dispense_media, \
            return_agar_plates, ref_kit_container

        p = Protocol()
        bacteria = p.ref(name="test_bact", id=None,
                         cont_type="micro-1.5", discard=True).well(0)
        media = return_dispense_media()["50_ug/ml_Kanamycin"]
        agar_id = return_agar_plates(1)[media]
        agar_plate = ref_kit_container(p, "my_agar_plate",
                                       "6-flat", agar_id, discard=False,
                                       store="cold_4")

    Parameters
    ----------
    protocol : Protocol
        instance of protocol.
    name : str
        Name for the plate.
    container : str
        Container type name.
    kit_id : str
        Kit item to be created.
    discard : bool
        Determine if plate is discarded after use.
    store : str
        If the plate is not discarded, indicate valid storage condition.

    Returns
    -------
    Container


    """
    kit_item = Container(None, protocol.container_type(
        container), name, storage=store if store else None)
    if store:
        protocol.refs[name] = Ref(
            name, {"reserve": kit_id, "store": {"where": store}}, kit_item)
    else:
        protocol.refs[name] = Ref(
            name, {"reserve": kit_id, "discard": discard}, kit_item)
    return kit_item


class ResourceIDs(object):

    """Common resource ids

    A list of resource identification numbers used to provision
    resources using Autoprotocol.

    Example
    -------

    .. code-block:: python

        from autoprotocol-utilities import ResourceIDs

        res = ResourceIDs()
        res.water
        res.te
        res.zymo_10b
        res.zymo_dh5a
        res.zymo_jm109
        res.ampicillin_100mg_ml
        res.chloramphenicol_34mg_ml
        res.kanamycin_50mg_ml
        res.mytaq_poly
        res.dpn_neb
        res.phusion_poly
        res.orange_g_100
        res.orange_g_500

    Returns
    -------
    str
        resource id of the reagent

    """

    def __init__(self):
        # diluent
        self.water = "rs17gmh5wafm5p"
        self.te = "rs17pwyc754v9t"
        # competent cells
        self.zymo_10b = "rs16pbjc4r7vvz"
        self.zymo_dh5a = "rs16pbj944fnny"
        self.zymo_jm109 = "rs16pbjdhwkjxy"
        # antibiotics
        self.ampicillin_100mg_ml = "rs17msfk8ujkca"
        self.chloramphenicol_34mg_ml = "rs17p6t8ty2ny4"
        self.kanamycin_50mg_ml = "rs17msfpgpbqyv"
        # media
        self.lb_miller_50ug_ml_kan = "rs18s8x88zz9ee"
        self.lb_miller_100ug_ml_amp = "rs18s8x4qbsvjz"
        self.lb_miller_noAB = "rs17bafcbmyrmh"
        self.tb_50ug_ml_kan = "rs18xqzy4ftdy3"
        self.tb_100ug_ml_amp = "rs18xr22jq7vtz"
        self.tb_100ug_ml_spec = "rs18xr25bpakgs"
        self.tb_25ug_ml_cm = "rs18xr28t3z8nx"
        # kunkel resources
        self.t7_poly = "rs16pca2urcz74"
        self.t4_pnk = "rs16pc9rd5hsf6"
        self.t4_pnk_buffer = "rs16pc9rd5sg5d"
        self.atp_100mM = "rs16pccshb6cb4"
        # SYBR green qPCR enzyme
        self.sensifast = "rs17knkh7526ha"
        self.itaq = "rs18fabf4h5se8"
        # mytaq dna polymerase resources
        self.mytaq_poly = "rs16pcbhquhaz3"
        self.mytaq_red_poly = "rs16r3h6umutcg"
        self.mytaq_hs_red_mix = "rs17kj4vnh5xm3"
        self.mytaq_red_mix = "rs17kj4j9vgh4x"
        self.mytaq_buffer = "rs16pcbhqurzpa"
        # phusion dna polymerase resources
        self.phusion_poly = "rs16pcc3mgay64"
        self.phusion_mgcl_50 = "rs16pcc3mgzy4r"
        self.phusion_gc_buffer = "rs16pcc3mgs9eh"
        self.phusion_hf_buffer = "rs16pcc3mgjnub"
        # kapa dna polymerase master mixes
        self.kapa_hifi_hs_mix = "rs18esg5hz25cm"
        self.kapa_2g_hs_mix = "rs18esfy3xqvut"
        # velocity dna polymerase resources
        self.velocity_poly = "rs16pcckjgaxdt"
        self.velocity_hifi_buffer = "rs16pcckjghhyz"
        self.velocity_mgcl = "rs16pcckjgs8p8"
        self.velocity_dmso = "rs16pcckjgyu9e"
        # la taq dna polymerase resources
        self.lataq_poly = "rs16pcbdc5dfw6"
        self.lataq_poly_gc1 = "rs16pcb53zwxjz"
        self.la_buffer_2_10x = "rs16pcbdc5n6kd"
        self.lataq_poly_gc = "rs16pcb53zp8vs"
        self.lataq_poly_gc2 = "rs16pcb5425j67"
        self.lataq_dntp25 = "rs16pcbdc5us6k"
        # pcr components
        self.dntps_25 = "rs16pcb542c5rd"
        self.dntps_10 = "rs186wj7fvknsr"
        self.mgcl = "rs16pca93rjwgq"
        self.dmso = "rs186hr8m38ntw"
        # restriction enzymes
        #  buffers
        self.cutsmart_buffer = "rs17ta93g3y85t"
        self.neb21_buffer = "rs17sh6krrzjqu"
        self.neb31_buffer = "rs18jwyqebdsdu"
        self.fastdigest_buffer = "rs18a8uvv7us8t"
        #  enzymes
        self.ncoi_hf = "rs183kfrjt4svz"
        self.psti_hf = "rs17usrub943jf"
        self.ecori_hf = "rs17ta8xftpdk6"
        self.bamhi_hf = "rs17ta8tz5ffby"
        self.bbsi = "rs17rrdaz88sz5"
        self.bsmbi = "rs17px7f9yg3kn"
        self.pvuii_hf = "rs17ta9v88fgpd"
        self.bsai = "rs17px78jjr2fq"
        self.ndei = "rs186h3y9nuqzb"
        self.xhoi = "rs186h4bcgjtyu"
        self.dpni_neb = "rs18kfcmf5xvxz"
        self.mfei_hf = "rs18nw6ta6d5bn"
        self.esp3i = "rs18a8ttpm8hxk"
        self.hindiii_hf = "rs18nw6kpnp44v"
        self.sali = "rs18trptum9gc4"
        self.smai = "rs18vvr4tgrghh"
        self.xbai = "rs18x6ja5k75ev"
        self.hincii = "rs18x6jrxfxtut"
        self.bbvCi = "rs18x6k25qmr6k"
        # orange g
        self.orange_g_100 = "rs17zw9zsaqd55"
        self.organge_g_500 = "rs17zwe6rux5b7"
        # control plasmids
        self.control_amp = "rs18rx59spw2t8"
        self.control_kan = "rs18rx6a44qss7"
        # ligases
        self.thermo_t4ligase_buffer = "rs16pc8u4dmsbg"
        self.thermo_t4ligase = "rs16pc8u4dd3n9"
        self.neb_t4ligase_buffer = "rs17sh5rzz79ct"
        self.neb_t4ligase = "rs16pc8krr6ag7"
        self.ligase_control = "rs18sfjf96tkwe"
        # other
        self.exosap = "rs18dnrskds4t6"
        # assembly reagents
        self.nebuilder2x = "rs18pc86ykcep6"
        self.nebuilderpc = "rs192pqa2jua9v"
        self.gibson2x = "rs16pfatkggmk5"
        self.gibsonpc = "rs1959tbv27xu2"
        self.infusion5x = "rs16pfv7qw5ytj"
        self.infusionpuc = "rs192pqw2nuef8"
        self.infusioninsert = "rs192pqxnx9gm2"
        # QuantIt
        self.quantItLambda = "rs18qstca8ksrt"
        self.quantItTE = "rs18qst9znacdy"
        self.quantItPico = "rs18qst83met3g"
        # MagJet
        self.lysozyme = "rs18u5sv3y8haj"
        self.dtt = "rs18umvdgu69su"
        self.MagJETRNALysisBuffer = "rs18umvv99scva"
        self.MagJETRNABeads = "rs18umvxjtubpw"
        self.MagJETRNAReactionBuffer = "rs18umwewhmmvr"
        self.MagJETRNADNase = "rs18umwj7mmmdc"
        # genotyping lysis buffers
        self.geno_lysis = "rs17krffpwfyrq"
        self.geno_neut = "rs17krfgz55nqq"

    def bacteria(self, bact=None):
        """Return competent bacteria id

        Parameters
        ----------
        bact: string
            Bacteria name, one of: 'Zymo 10B', 'Zymo DH5a', 'Zymo JM109'

        Returns
        -------
        string
            resource id for the bacteria requested, `None` if bact could not
            be found
        """
        bacteria = {"Zymo 10B": self.zymo_10b,
                    "Zymo DH5a": self.zymo_dh5a,
                    "Zymo JM109": self.zymo_jm109}
        return bacteria.get(bact)

    def diluents(self, dil=None):
        """Return diluent id

        Parameters
        ----------
        dil: string
            Diluent name, one of: 'water', 'TE'

        Returns
        -------
        string
            resource id for the diluent requested, `None` if dil could not
            be found
        """
        diluents = {"water": self.water,
                    "TE": self.te}
        return diluents.get(dil)

    def exoassembly_kits(self, kit=None):
        kit_dict = {"NEBuilder": {"name": "NEBuilder",
                                  "dil_fact": 2,
                                  "resource": self.nebuilder2x,
                                  "pc": {self.nebuilderpc: 2}},
                    "Gibson": {"name": "Gibson",
                               "dil_fact": 2,
                               "resource": self.gibson2x,
                               "pc": {self.gibsonpc: 2}},
                    "InFusion": {"name": "InFusion",
                                 "dil_fact": 5,
                                 "resource": self.infusion5x,
                                 "pc": {self.infusionpuc: Unit(1, "uL"),
                                        self.infusioninsert: Unit(2, "uL")}}}
        return kit_dict.get(kit)

    def transformation_controls(self, media=None):
        """Return transformation controls

        Parameters
        ----------
        media: string
            Media for which to select the control. One of:
            'lb_miller_50ug_ml_kan', 'lb_miller_100ug_ml_amp'

        Returns
        -------
        string
            resource id for the positive control requested, `None` if media
            could not be found
        """
        controls = {"lb_miller_50ug_ml_kan": self.control_kan,
                    "lb_miller_100ug_ml_amp": self.control_amp}
        return controls.get(media)

    def growth_media(self, media=None):
        """Return growth media resource id

        Parameters
        ----------
        media: string
            Media for which to select the media.

        Returns
        -------
        string
            resource id for the media requested, `None` if media
            could not be found
        """
        media_dict = {
            "lb_miller_50ug_ml_kan": self.lb_miller_50ug_ml_kan,
            "lb_miller_100ug_ml_amp": self.lb_miller_100ug_ml_amp,
            "lb_miller_noAB": self.lb_miller_noAB,
            "tb_50ug_ml_kan": self.tb_50ug_ml_kan,
            "tb_100ug_ml_amp": self.tb_100ug_ml_amp,
            "tb_100ug_ml_specto": self.tb_100ug_ml_spec,
            "tb_25ug_ml_cm": self.tb_25ug_ml_cm,
        }
        return media_dict.get(media)

    def t4_ligase(self, ligase_type=None):
        """Return T4 ligase reagents

        Parameters
        ----------
        media: string
            Vendor to use. 'neb' or 'thermo'.

        Returns
        -------
        dict
            'buffer' - resource id for the buffer
            'ligase' - resource id for the ligase

        """
        ligases = {"neb": {"buffer": self.neb_t4ligase_buffer,
                           "ligase": self.neb_t4ligase},
                   "thermo": {"buffer": self.thermo_t4ligase_buffer,
                              "ligase": self.thermo_t4ligase}}
        return ligases.get(ligase_type)

    def restriction_enzyme_buffers(self, enzyme):
        """Returns a tuple of enzyme_id and buffer_id for a given enzyme

        Restriction enzymes need to be digested with a certain buffer.
        This function returns the correct resource id for the enzyme as well
        as for the matching buffer.

        Parameters
        ----------
        enzyme: string
            name of enzyme, all lower caps such as `dpni_neb` or `ncoi_hf`

        Returns
        -------
        namedtuple
            enzyme_id, buffer_id, errors
        """
        em = []
        resset = namedtuple('RestrictionSet', 'enzyme_id buffer_id errors')

        enzymes = {"ncoi_hf": self.ncoi_hf,
                   "psti_hf": self.psti_hf,
                   "ecori_hf": self.ecori_hf,
                   "bamhi_hf": self.bamhi_hf,
                   "bbsi": self.bbsi,
                   "bsmbi": self.bsmbi,
                   "pvuii_hf": self.pvuii_hf,
                   "bsai": self.bsai,
                   "ndei": self.ndei,
                   "xhoi": self.xhoi,
                   "dpni_neb": self.dpni_neb,
                   "mfei_hf": self.mfei_hf,
                   "esp3i": self.esp3i,
                   "hindiii_hf": self.hindiii_hf,
                   "sali": self.sali,
                   "smai": self.smai,
                   "bbvCi": self.bbvCi,
                   "xbai": self.xbai,
                   "hincii": self.hincii}

        buffer_map = {}
        buffer_map["cutsmart"] = {
            "buffer_id": self.cutsmart_buffer, "enzymes": [
                "pvuii_hf", "ecori_hf", "hindiii_hf",
                "bamhi_hf", "psti_hf", "ncoi_hf", "xbai",
                "bsai", "ndei", "xhoi", "smai", "bbvCi",
                "dpni_neb", "mfei_hf"]}
        buffer_map["neb_21"] = {
            "buffer_id": self.neb21_buffer, "enzymes": ["bbsi"]}
        buffer_map["new_31"] = {
            "buffer_id": self.neb31_buffer, "enzymes": [
                "bsmbi", "sali", "hincii"]}
        buffer_map["fast_digest"] = {
            "buffer_id": self.fastdigest_buffer, "enzymes": ["esp3i"]}

        enzyme_id = enzymes.get(enzyme, None)
        if not enzyme_id:
            em.append("The enzyme (%s) cannot be found." % enzyme)
        buffer_id = None
        for resbuffer in buffer_map.values():
            if enzyme in resbuffer["enzymes"]:
                buffer_id = resbuffer["buffer_id"]
        if not buffer_id:
            em.append("The enzyme specified (%s) doesn't have a corresponding"
                      " buffer." % enzyme)

        em = [_f for _f in em if _f]
        if len(em) == 0:
            em = None

        return resset(enzyme_id=enzyme_id, buffer_id=buffer_id, errors=em)



from autoprotocol.unit import Unit
import sys

if sys.version_info[0] >= 3:
    string_type = str
else:
    string_type = basestring


def melt_curve(start=65, end=95, inc=0.5, rate=5):
    """Generate a melt curve on the fly

    No inputs neded for a standard melt curve.

    Example Usage:

    .. code-block:: python

        from autoprotocol import Protocol
        from autoprotocol_utilities import melt_curve
        protocol = Protocol()

        dest_plate = protocol.ref("plate", None, "96-pcr", discard=True,
                                  storage=None)
        protocol.seal(dest_plate)
        melt_params = melt_curve()
        protocol.thermocycle(dest_plate,
                             [{"cycles": 1,
                               "steps": [
                                   {"temperature": "37:celsius",
                                   "duration": "60:minute"}
                                   ]
                               }
                              ],
                             volume="15:microliter",
                             dataref="data",
                             dyes={"SYBR":
                             dest_plate.wells_from(0, 3).indices()},
                             **melt_params)

    Returns:

    .. code-block:: json

        {
          "refs": {
            "plate": {
              "new": "96-pcr",
              "discard": true
            }
          },
          "instructions": [
            {
              "object": "plate",
              "type": "ultra-clear",
              "op": "seal"
            },
            {
              "dataref": "data",
              "melting": {
                "start": "65.00:celsius",
                "rate": "5.00:second",
                "end": "95.00:celsius",
                "increment": "0.50:celsius"
              },
              "object": "plate",
              "volume": "15:microliter",
              "groups": [
                {
                  "cycles": 1,
                  "steps": [
                    {
                      "duration": "60:minute",
                      "temperature": "37:celsius"
                    }
                  ]
                }
              ],
              "dyes": {
                "SYBR": [
                  "A1",
                  "A2",
                  "A3"
                ]
              },
              "op": "thermocycle"
            }
          ]
        }

    Parameters
    ----------
    start : int, float
        Temperature to start at
    end : int, float
        Temperature to end at
    inc : int, float
        Temperature increment during the melt_curve
    rate : int
        After x seconds the temperature is incremented by inc

    Returns
    -------
    melt_params : dict
        containing melt_params

    Raises
    ------
    ValueError
        If start, end or inc are not of type `float` or `int`. And if rate is
        not of type `int`

    """
    assert isinstance(start, (float, int))
    assert isinstance(end, (float, int))
    assert isinstance(inc, (float, int))
    assert isinstance(rate, int)

    melt_params = {"melting_start": "%.2f:celsius" % start,
                   "melting_end": "%.2f:celsius" % end,
                   "melting_increment": "%.2f:celsius" % inc,
                   "melting_rate": "%.2f:second" % rate}
    return melt_params


def thermocycle_ramp(start_temp, end_temp, total_duration, step_duration):
    """Create a ramp instruction for the thermocyler.

    Create a multi-temperature thermocycling program commonly used in
    annealing protocols. Based on total time and the step duration this
    function computes the temperature increment required for each step within
    the start and the end temperature.

    Example Usage:

    .. code-block:: python

        thermocycle_group = [
          {
              "cycles": 1,
              "steps": thermocycle_ramp(65, 95, "30:minute", "1:minute")
          }
        ]
        protocol.thermocycle(dest_plate,
                             groups=thermocycle_group,
                             volume="15:microliter")

    Parameters
    ----------
    start_temp: string, int, float, Unit
        Start of the thermocycle protocol, in the format "37:celsius"
    end_temp: string, int, float, Unit
        End of the thermocycle protocol, in the format "37:celsius"
    total_duration: string, Unit
        Total duration of the thermocycle protocol, in the format "1:hour"
    step_duration: string, Unit
        Time that each temperature should be held, in the format "1:minute"

    Returns
    -------
    dict
        containing thermocycling steps that can be used in the
        thermocycle instruction

    Raises
    ------
    ValueError
        If either temperature is not of type `int`, `float`, `string` or
        `Unit` and if either duration is not of type `string` or `Unit`

    """
    assert isinstance(start_temp, (int, float, string_type, Unit))
    assert isinstance(end_temp, (int, float, string_type, Unit))
    assert isinstance(total_duration, (string_type, Unit))
    assert isinstance(step_duration, (string_type, Unit))

    if isinstance(start_temp, string_type):
        start_temp = Unit.fromstring(start_temp)
    elif isinstance(start_temp, (int, float)):
        start_temp = Unit(start_temp, 'degC')
    if isinstance(end_temp, string_type):
        end_temp = Unit.fromstring(end_temp)
    elif isinstance(end_temp, (int, float)):
        end_temp = Unit(end_temp, 'degC')
    if isinstance(total_duration, string_type):
        total_duration = Unit.fromstring(total_duration)
    if isinstance(step_duration, string_type):
        step_duration = Unit.fromstring(step_duration)

    start_temp.to('degC')
    end_temp.to('degC')
    total_duration.to_base_units()
    step_duration.to_base_units()

    num_steps = int(total_duration // step_duration)
    step_size = (end_temp - start_temp).magnitude / num_steps

    thermocycle_steps = []
    for i in range(num_steps + 1):
        thermocycle_steps.append({
            "temperature": "%s:celsius" % (
                start_temp.magnitude + i * step_size),
            "duration": str(step_duration)
        })
    return thermocycle_steps



from autoprotocol.unit import Unit


def dna_mass_to_mole(length, mass, ds=True):
    """
    For the DNA Length and mass given, return the mole amount of DNA

    Example Usage:

    .. code-block:: python

        from autoprotocol_utilities import dna_mass_to_mole
        from autoprotocol.unit import Unit

        dna_length = 100
        dna_mass = Unit(33, 'ng')
        dna_mass_to_mole(dna_length, dna_mass)

    Returns:

    .. code-block:: python

        Unit(0.5, 'picomole')

    Parameters
    ----------
    length: int
        Length of DNA in bp
    mass: str, Unit
        Weight of DNA in prefix-g
    ds: bool, optional
        True for dsDNA, False for ssDNA

    Returns
    -------
    pmole_dna: Unit
        Mole amount of DNA in pmol

    Raises
    ------
    ValueError
        If inputs are not of specified types

    """
    if isinstance(mass, str):
        mass = Unit.fromstring(mass)

    if not isinstance(mass, Unit) or str(mass.dimensionality) != "[mass]":
        raise ValueError("Mass of DNA must be of type Unit in prefix-gram")

    if not isinstance(length, int):
        raise ValueError(
            "Length of DNA is of type %s, must be of type "
            "integer" % type(length))

    if not isinstance(ds, bool):
        raise ValueError(
            "ds is of type %s, must be of type bool: True for dsDNA, "
            "False for ssDNA" % type(ds))

    dna_pg = mass.to("pg")

    if ds:
        dna_pmol = dna_pg / (Unit(660, "pg/pmol") * length)
    else:
        dna_pmol = dna_pg / (Unit(330, "pg/pmol") * length)

    return dna_pmol


def dna_mole_to_mass(length, mole, ds=True):
    """
    For the DNA Length and mole amount given, return the mass of DNA

    Example Usage:

    .. code-block:: python

        from autoprotocol_utilities import dna_mole_to_mass
        from autoprotocol.unit import Unit

        dna_length = 5000
        dna_mole = "10:pmol"
        dna_mole_to_mass(dna_length, dna_mole)

    Returns:

    .. code-block:: python

        Unit(33.0, 'microgram')

    Parameters
    ----------
    length: int
        Length of DNA in bp
    mole: str, Unit
        Mole amount of DNA in prefix-mol
    ds: bool, optional
        True for dsDNA, False for ssDNA

    Returns
    -------
    dna_ug: Unit
        Weight of DNA in ug

    Raises
    ------
    ValueError
        If inputs are not of specified types

    """
    if isinstance(mole, str):
        mole = Unit.fromstring(mole)

    if not isinstance(mole, Unit) or str(mole.dimensionality) != "[substance]":
        raise ValueError(
            "Mole amount of DNA must be of type Unit in prefix-mol")

    if not isinstance(length, int):
        raise ValueError(
            "Length of DNA is of type %s, must be of type "
            "integer" % type(length))

    if not isinstance(ds, bool):
        raise ValueError(
            "ds is of type %s, must be of type bool: True for dsDNA, "
            "False for ssDNA" % type(ds))

    dna_pmol = mole.to("pmol")

    if ds:
        dna_ug = (
            Unit(660, "pg/pmol") * dna_pmol * Unit(10**(-6), "ug/pg") * length)
    else:
        dna_ug = (
            Unit(330, "pg/pmol") * dna_pmol * Unit(10**(-6), "ug/pg") * length)

    return dna_ug


def molar_to_mass_conc(length, molar, ds=True):
    """
    For the DNA molarity given, return the mass concentration of DNA

    Example Usage:

    .. code-block:: python

        from autoprotocol_utilities import molar_to_mass_conc
        from autoprotocol_utilities import dna_mole_to_mass
        from autoprotocol.unit import Unit

        dna_length = 5000
        dna_molarity = Unit(10, 'uM')
        molar_to_mass_conc(dna_length, dna_molarity)

    Returns:

    .. code-block:: python

        Unit(33000.0, 'nanogram / microliter')

    Parameters
    ----------
    length: int
        Length of DNA in bp
    molar: str, Unit
        Molarity of DNA in prefix-M
    ds: bool, optional
        True for dsDNA, False for ssDNA

    Returns
    -------
    mass_conc: Unit
        Mass concentration of DNA in ng/uL

    Raises
    ------
    ValueError
        If inputs are not of specified types

    """
    if not isinstance(length, int):
        raise ValueError(
            "Length of DNA is of type %s, must be of type "
            "integer" % type(length))

    if isinstance(molar, str):
        molar = Unit.fromstring(molar)

    if not (isinstance(molar, Unit) and
            str(molar.dimensionality) == '[substance] / [length] ** 3'):
        raise ValueError(
            "Molar concentration of DNA must be of type string or Unit")

    if not isinstance(ds, bool):
        raise ValueError(
            "ds is of type %s, must be of type bool: True for dsDNA, "
            "False for ssDNA" % type(ds))

    dna_umole = Unit((molar / Unit(1, "M")).magnitude, "umol")
    dna_ug = dna_mole_to_mass(length, dna_umole, ds)
    mass_conc = Unit(dna_ug.magnitude * 1000, "ng/uL")

    return mass_conc


def mass_conc_to_molar(length, mass_conc, ds=True):
    """
    For the DNA mass concentration given, return the molarity of DNA

    Example Usage:

    .. code-block:: python

        from autoprotocol_utilities import mass_conc_to_molar
        from autoprotocol_utilities import dna_mass_to_mole
        from autoprotocol.unit import Unit

        dna_length = 5000
        dna_mass_conc = Unit(33, 'ng/uL')
        mass_conc_to_molar(dna_length, dna_mass_conc)

    Returns:

    .. code-block:: python

        Unit(0.01, 'micromolar')

    Parameters
    ----------
    length: int
        Length of DNA in bp
    mass_conc: str, Unit
        Mass concentration of DNA
    ds: bool, optional
        True for dsDNA, False for ssDNA

    Returns
    -------
    molar: Unit
        Molarity of DNA in uM

    Raises
    ------
    ValueError
        If inputs are not of specified types

    """
    if not isinstance(length, int):
        raise ValueError(
            "Length of DNA is of type %s, must be of type "
            "integer" % type(length))

    if isinstance(mass_conc, str):
        mass_conc = Unit.fromstring(mass_conc)

    if not isinstance(mass_conc, Unit) or \
            str(mass_conc.dimensionality) != '[mass] / [length] ** 3':
        raise ValueError("Mass concentration of DNA must be of type Unit")

    if not isinstance(ds, bool):
        raise ValueError(
            "ds is of type %s, must be of type bool: True for dsDNA, "
            "False for ssDNA" % type(ds))

    dna_ng = Unit((mass_conc / Unit(1, "ng/uL")).magnitude, "ng")
    dna_pmol = dna_mass_to_mole(length, dna_ng, ds)
    dna_molar = Unit(round(dna_pmol.magnitude, 9), "uM")

    return dna_molar


def ligation_insert_ng(plasmid_size,  plasmid_mass,
                       insert_size, molar_ratio=1):
    """
    For the plasmid size, plasmid amount, insert size, and molar ratio given,
    return the mass of insert needed for ligation

    Different from ligation_insert_volume: no insert concentration is
    given -> returns mass of insert needed

    Example Usage:

    .. code-block:: python

        from autoprotocol_utilities import ligation_insert_ng
        from autoprotocol.unit import Unit

        plasmid_size = 3000
        plasmid_mass = Unit(100, 'ng')
        insert_size = 48
        ligation_insert_ng(plasmid_size, plasmid_mass, insert_size)

    Returns:

    .. code-block:: python

        Unit(1.6, 'nanogram')

    Parameters
    ----------
    plasmid_size : int
        Length of plasmid in bp.
    insert_size: int
        Length of insert in bp
    plasmid_mass : str, Unit
        Mass of plasmid in prefix-g
    molar_ratio : int, float, string, optional
        Ligation molar ratio of insert : vector.  By default it is 1 : 1.
        Generally ligations are tested at 1:3, 1:1, and 3:1

    Returns
    -------
    insert_amount: Unit
        Amount of insert solution needed in ng

    Raises
    ------
    ValueError
        If wells are not of type list, WellGroup or Container

    """

    # Check input types
    if not isinstance(plasmid_size, int):
        raise ValueError("Plasmid_size: must be an integer")

    if not isinstance(insert_size, int):
        raise ValueError("insert_size: must be an integer")

    if type(molar_ratio) == str:
        molar_ratio = float(
            molar_ratio.split(":")[0]) / float(molar_ratio.split(":")[1])

    if type(molar_ratio) not in (int, float):
        raise ValueError(
            "molar_ratio: must be an int, float, or string in the form "
            "of int:int")

    if isinstance(plasmid_mass, str):
        plasmid_mass = Unit.fromstring(plasmid_mass)

    if not (isinstance(plasmid_mass, Unit) and
            str(plasmid_mass.dimensionality) == "[mass]"):
        raise ValueError(
            "Plasmid amount must be of type str or Unit in prefix-g")

    length_ratio = float(insert_size) / float(plasmid_size)
    plasmid_ng = plasmid_mass.to("ng")
    insert_ng = plasmid_ng * length_ratio * molar_ratio

    return insert_ng


def ligation_insert_volume(plasmid_size,  plasmid_mass, insert_size,
                           insert_conc, ds=True, molar_ratio=1):
    """
    For the plasmid size, plasmid amount, insert size, insert concentration,
    and molar ratio given, return the volume of insert solution needed for
    ligation

    Different from ligation_insert_ng: insert concentration is given -> returns
    volume of insert solution needed

    Example Usage:

    .. code-block:: python

        from autoprotocol_utilities import ligation_insert_volume
        from autoprotocol_utilities import molar_to_mass_conc
        from autoprotocol.unit import Unit

        plasmid_size = 3000
        plasmid_mass = Unit(100, 'ng')
        insert_size = 48
        insert_conc = Unit(25, 'ng/uL')
        ligation_insert_volume(plasmid_size, plasmid_mass, insert_size,
                               insert_conc)

    Returns:

    .. code-block:: python

        Unit(0.064, 'microliter')

    Parameters
    ----------
    plasmid_size : int
        Length of plasmid in bp.
    plasmid_mass : str, Unit
        Mass of plasmid in prefix-g
    insert_size: int
        Length of insert in bp
    insert_conc: str, Unit
        Molar or mass concentration of insert
    ds: bool, optional
        True for dsDNA, False for ssDNA
    molar_ratio : int, float, string, optional
        Ligation molar ratio of insert : vector.
        Common ratios are 1:3, 1:1, and 3:1. 1:1 by default

    Returns
    -------
    insert_amount: Unit
        Volume of insert solution needed in uL

    Raises
    ------
    ValueError
        If wells are not of type list, WellGroup or Container

    """

    conc_dimension = ["[substance] / [length] ** 3", '[mass] / [length] ** 3']

    # Check input types
    if not isinstance(plasmid_size, int):
        raise ValueError("Plasmid_size: must be an integer")

    if isinstance(plasmid_mass, str):
        plasmid_mass = Unit.fromstring(plasmid_mass)

    if not isinstance(plasmid_mass, Unit) and \
            str(plasmid_mass.dimensionality) == "[mass]":
        raise ValueError(
            "Plasmid mass must be of type str or Unit in prefix-g")

    if not isinstance(insert_size, int):
        raise ValueError("insert_size: must be an integer")

    if isinstance(insert_conc, str):
        insert_conc = Unit.fromstring(insert_conc)

    if not (isinstance(insert_conc, Unit) and
            str(insert_conc.dimensionality) in conc_dimension):
        raise ValueError(
            "Plasmid concentration must be of type Unit in prefix-M or "
            "prefix-g / prefix-L ")

    if not isinstance(ds, bool):
        raise ValueError(
            "ds is of type %s, must be of type bool: True for dsDNA, "
            "False for ssDNA" % type(ds))

    if type(molar_ratio) == str:
        molar_ratio = float(
            molar_ratio.split(":")[0]) / float(molar_ratio.split(":")[1])

    if type(molar_ratio) not in (int, float):
        raise ValueError(
            "molar_ratio: must be an int, float, or string in the "
            "form of int:int")

    len_ratio = float(insert_size) / float(plasmid_size)
    plasmid_ng = plasmid_mass.to("ng")
    insert_ng = plasmid_ng * len_ratio * molar_ratio

    # Convert concentration to ng/uL
    if str(insert_conc.dimensionality) == conc_dimension[0]:
        insert_conc = molar_to_mass_conc(insert_size, insert_conc, ds)

    else:
        insert_conc = insert_conc.to("ng/uL")

    insert_vol = insert_ng / insert_conc

    return insert_vol


def ligation_insert_amount(plasmid_size, plasmid_conc, plasmid_volume,
                           insert_size, insert_conc, ds=True, molar_ratio=1):
    """
    For the plasmid size, plasmid concentration, insert size, insert
    concentration, and molar ratio given,
    return the volume of insert solution needed for ligation

    Different form ligation_insert_volume: plasmid concentration and volume
    are given instead of plasmid mass

    Example Usage:

    .. code-block:: python

        from autoprotocol_utilities import ligation_insert_amount
        from autoprotocol_utilities import molar_to_mass_conc
        from autoprotocol.unit import Unit

        plasmid_size = 2000
        plasmid_conc = '1.5:uM'
        plasmid_volume = Unit(10, 'uL')
        insert_size = 25
        insert_conc = Unit(10, 'ng/uL')
        ligation_insert_amount(plasmid_size, plasmid_conc, plasmid_volume,
                               insert_size, insert_conc)
    Returns:

    .. code-block:: python

        Unit(24.75, 'microliter')

    Parameters
    ----------
    plasmid_size : int
        Length of plasmid in bp.
    plasmid_conc : str, Unit
        Molar or mass concentration of plasmid solution
    plasmid_volume: str, Unit
        Volume of plasmid solution in prefix-L
    insert_size: int
        Length of insert in bp
    insert_conc : str, Unit
        Molar or mass concentration of insert solution
    ds: bool, optional
        True for dsDNA, False for ssDNA
    molar_ratio : int, float, string, optional
        Ligation molar ratio of insert : vector.
        Common ratios are 1:3, 1:1, and 3:1. 1:1 by default

    Returns
    -------
    insert_amount: Unit
        Volume of insert solution in uL

    Raises
    ------
    ValueError
        If wells are not of type list, WellGroup or Container

    """
    # Check input types
    if not isinstance(plasmid_size, int):
        raise ValueError("Plasmid_size: must be an integer")

    if not isinstance(insert_size, int):
        raise ValueError("insert_size: must be an integer")

    if isinstance(plasmid_volume, str):
        plasmid_volume = Unit.fromstring(plasmid_volume)
    if not isinstance(plasmid_volume, Unit) or \
            str(plasmid_volume.dimensionality) != "[length] ** 3":
        raise ValueError(
            "Volume of plasmid solution must be of type str or Unit")

    conc_dimension = ["[substance] / [length] ** 3", '[mass] / [length] ** 3']
    conc = [plasmid_conc, insert_conc]
    size = [plasmid_size, insert_size]
    for i in range(0, 2):
        if isinstance(conc[i], str):
            conc[i] = Unit.fromstring(conc[i])
        if (isinstance(conc[i], Unit) and
                str(conc[i].dimensionality) in conc_dimension):
            # Convert all concentrations to ng/uL
            if str(conc[i].dimensionality) == conc_dimension[0]:
                conc[i] = molar_to_mass_conc(size[i], conc[i], ds)
            else:
                conc[i] = conc[i].to("ng/uL")
        else:
            raise ValueError(
                "Concentration must be of type string or Unit ")

    if not isinstance(ds, bool):
        raise ValueError(
            "ds is of type %s, must be of type bool: True for dsDNA, "
            "False for ssDNA" % type(ds))

    if type(molar_ratio) == str:
        molar_ratio = float(
            molar_ratio.split(":")[0]) / float(molar_ratio.split(":")[1])

    if type(molar_ratio) not in (int, float):
        raise ValueError(
            "molar_ratio: must be an int, float, or string in the "
            "form of int:int")

    plasmid_conc = conc[0]
    insert_conc = conc[1]
    # Convert input volume to uL
    plasmid_uL = Unit((plasmid_volume / Unit(1, "uL")).magnitude, "uL")
    len_ratio = float(insert_size) / float(plasmid_size)
    plasmid_ng = plasmid_conc * plasmid_uL
    insert_ng = plasmid_ng * len_ratio * molar_ratio
    insert_amount = insert_ng / insert_conc

    return insert_amount



from setuptools import setup

setup(name='autoprotocol_utilities',
      version='2.3.2',
      description='Helper methods to construct protocols using autoprotocol-python.',  # NOQA
      url='https://github.com/transcriptic/autoprotocol-utilities',
      author='All past and present Application Scientists',
      author_email='vanessa@transcriptic.com',
      license='MIT',
      packages=['autoprotocol_utilities'],
      tests_require=['pytest'],
      install_requires=['autoprotocol>=3.7'],
      zip_safe=False)
